# coding: utf-8
import unittest
import basico
import basico.model_info


class TestBasicoModelManipulation(unittest.TestCase):

    def setUp(self):
        self.model = basico.new_model(name='Test Model')
        self.assertTrue(self.model.getModel().getObjectName() == 'Test Model')

    def test_add_compartment(self):
        basico.add_compartment('c', 1.0)
        c = basico.as_dict(basico.get_compartments('c'))
        self.assertAlmostEqual(c['initial_size'], 1)
        basico.set_compartment('c', initial_size=2)
        c = basico.as_dict(basico.get_compartments('c'))
        self.assertAlmostEqual(c['initial_size'], 2)

    def test_add_species(self):
        basico.add_species('X')
        s = basico.as_dict(basico.get_species('X'))
        self.assertAlmostEqual(s['initial_concentration'], 1)
        c = basico.as_dict(basico.get_compartments('compartment'))
        self.assertAlmostEqual(c['initial_size'], 1)
        basico.set_species('X', initial_concentration=2)
        s = basico.as_dict(basico.get_species('X'))
        self.assertAlmostEqual(s['initial_concentration'], 2)

    def test_add_event(self):
        basico.add_event('e0', 'Time > 10', [['X', '0']])
        e = basico.get_events('e0')
        self.assertEqual(str(e.iloc[0]['trigger']), 'Time > 10')

    def test_add_reaction(self):
        basico.add_reaction('r0', 'X -> Y')
        r = basico.get_reactions('r0')
        self.assertEqual(str(r.iloc[0]['scheme']), 'X -> Y')
        params = basico.get_reaction_parameters(reaction_name='r0')
        self.assertEqual(params.loc['(r0).k1'].value, 0.1)
        basico.set_reaction_parameters('(r0).k1', value=1)
        params = basico.get_reaction_parameters(reaction_name='r0')
        self.assertEqual(params.loc['(r0).k1'].value, 1)

    def test_add_parameter(self):
        basico.add_parameter('sin_time', type='assignment', expression='sin({Time})')
        p = basico.get_parameters('sin_time')
        self.assertEqual('assignment', str(p.iloc[0]['type']))
        self.assertEqual('sin ( Time )', str(p.iloc[0]['expression']))
        basico.remove_parameter('sin_time')

    def test_set_annotation(self):
        if not basico.have_miriam_resources():
            basico.update_miriam_resources()

        basico.set_miriam_annotation(
            creators=[
                {
                    'first_name': 'Frank',
                    'last_name': 'Bergmann',
                    'email': 'fbergman@caltech.edu',
                    'organization': 'Heidelberg University'}
            ],

        )

        self.assertTrue(basico.have_miriam_resources())

        annotations = basico.get_miriam_annotation()
        self.assertEqual(len(annotations['creators']), 1)
        self.assertEqual(annotations['creators'][0]['first_name'], 'Frank')
        self.assertEqual(annotations['creators'][0]['last_name'], 'Bergmann')
        self.assertEqual(annotations['creators'][0]['email'], 'fbergman@caltech.edu')
        self.assertEqual(annotations['creators'][0]['organization'], 'Heidelberg University')

        basico.set_miriam_annotation(descriptions=[
          {
            'qualifier': 'is',
            'resource': 'BioModels Database',
            'id': 'MODEL6623915522',
          }
        ])

        annotations = basico.get_miriam_annotation()
        self.assertEqual(len(annotations['descriptions']), 1)
        self.assertEqual(annotations['descriptions'][0]['qualifier'], 'is')
        self.assertEqual(annotations['descriptions'][0]['resource'], 'BioModels Database')
        self.assertEqual(annotations['descriptions'][0]['id'], 'MODEL6623915522')

        basico.set_miriam_annotation(references=[
          {
            'id': '10951190',
            'resource': 'PubMed',
          }
        ])

        annotations = basico.get_miriam_annotation()
        self.assertEqual(len(annotations['references']), 1)
        self.assertEqual(annotations['references'][0]['id'], '10951190')
        self.assertEqual(annotations['references'][0]['resource'], 'PubMed')

    def test_add_ode(self):
        basico.add_equation('d[PG]/dt=([E]/((k_4 [7-ADCA])/K_n +(k_5 [7-ADCA])/K_n +(k_6 [PGME])/K_si +k3))((k_2 [PGME])/K_s   +(k_4b [CEX])/K_p )(k_3+(k_5 [7-ADCA])/K_n   +(k_6 [PGME])/K_si )')
        basico.add_equation(
            'd[CEX]/dt=(k_2 [E][PGME])/K_s -([E]/((k_4 [7-ADCA])/K_n +(k_5 [7-ADCA])/K_n +(k_6 [PGME])/K_si +k3))((k_2 [PGME])/K_s   +(k_4b [CEX])/K_p )(k_3+(k_5 [7-ADCA])/K_n )')
        basico.add_equation(
            '[E]=([EA]_0 exp(-k_d t))/(1+([PGME])/K_s + ((k_2 [PGME])/(K_s ((k_4 [7-ADCA])/K_n +(k_5 [7-ADCA])/K_n +(k_6 [PGME])/K_si +k3) ))(1+([7-ADCA])/K_n +([PGME])/K_si )+([CEX])/K_p +([PG])/K_p2i )')
        basico.add_equation('d[S1]/dt=-r_s1*w_1*[B]')
        species = basico.get_species()
        self.assertEqual(species.loc['PG'].type, 'ode')
        self.assertEqual(species.loc['S1'].type, 'ode')
        self.assertEqual(species.loc['PG'].expression, '[E] / ( Values[k_4] * [7-ADCA] / Values[K_n] + Values[k_5] * [7-ADCA] / Values[K_n] + Values[k_6] * [PGME] / Values[K_si] + Values[k3] ) * ( Values[k_2] * [PGME] / Values[K_s] + Values[k_4b] * [CEX] / Values[K_p] ) * ( Values[k_3] + Values[k_5] * [7-ADCA] / Values[K_n] + Values[k_6] * [PGME] / Values[K_si] )')
        self.assertEqual(species.loc['E'].type, 'assignment')
        self.assertEqual(species.loc['E'].expression, '[EA]_0 * exp ( - Values[k_d] * Time ) / ( 1 + [PGME] / Values[K_s] + Values[k_2] * [PGME] / ( Values[K_s] * ( Values[k_4] * [7-ADCA] / Values[K_n] + Values[k_5] * [7-ADCA] / Values[K_n] + Values[k_6] * [PGME] / Values[K_si] + Values[k3] ) ) * ( 1 + [7-ADCA] / Values[K_n] + [PGME] / Values[K_si] ) + [CEX] / Values[K_p] + [PG] / Values[K_p2i] )')

        basico.add_equation('obs=sin(t)')
        self.assertEqual(basico.get_parameters().loc['obs'].type, 'assignment')
        self.assertEqual(basico.get_parameters().loc['obs'].expression, 'sin ( Time )')

        # encountered an odd issue when pasting in equations with unicode characters
        tokens = basico.model_info._tokenize_eqn(u'[E]=([E]_0  exp⁡(-k_d t))/(1+[PGME]/K_s +[PGME]^2/(K_s K_si )+[7-ADCA]/K_n +((k_2 [PGME])/(K_s ((k_4 [7-ADCA])/K_n +(k_5 [7-ADCA])/K_n +k3) ))((k_2 [PGME])/K_s   +(k_7 [PGME]^2)/(K_s K_si )+(k_4b [CEX])/K_p )(1+[7-ADCA]/K_n +[PGME]/K_si )+[CEX]/K_p +[PG]/K_p2i )')
        self.assertTrue(tokens is not None)
        self.assertEqual(tokens['tokens'][3], 'exp')

    def test_add_function(self):
        # add
        basico.add_function('MA KD', 'vr*(K_d-1)')
        funs = basico.get_functions('MA KD')
        self.assertTrue(funs is not None)
        fun = funs.iloc[0]
        self.assertEqual(fun.reversible, False)
        self.assertEqual(fun.formula, 'vr*(K_d-1)')
        self.assertEqual(fun.general, True)
        # remove
        basico.remove_function('MA KD')
        funs = basico.get_functions('MA KD')
        self.assertTrue(funs is None)

    def test_change_mapping(self):
        basico.add_function('MA Keq', 'vr*(Keq-1)')
        basico.add_parameter('K_d', initial_value=1)
        basico.add_reaction('r1 ma keq', scheme='A -> B', function='MA KD', mapping={'vr': 1000, 'Keq': 'K_d'})

    def test_expressions(self):
        basico.new_model(name='Model 4', volume_unit='ml', quantity_unit='mmol')

        basico.add_species('PGME', initial_concentration=20)
        basico.add_species('7-ADCA', initial_concentration=10)
        basico.add_species('CEX', initial_concentration=0)
        basico.add_species('PG', initial_concentration=2)

        params = ['K_s', 'k2',  'K_n', 'k3',  'k4',  'k5',  'k6',   'K_si', 'K_p', 'k_4b',  'e0', 'kd', 'K_p2i']
        values = [14,25920,290,25020,4416000,29460,99600,20,39,547560,0.000198000000000000,0.0385800000000000,12]
        for k, v in zip(params, values):
            basico.add_parameter(k, initial_value=v)

        basico.add_equation('A = k4 * [7-ADCA] / K_n + k5 * [7-ADCA] / K_n + k6 * [PGME] / K_si + k3')
        basico.add_equation('e = e0 /(1 + [PGME]^2/ (K_s * K_si) + [7-ADCA] / K_n + ( k2 * [PGME] /(K_s * A))*( 1 + [7-ADCA]/K_n + [PGME]/K_si)+[CEX]/K_p+[PG]/K_p2i)*exp(-kd*t)')
        basico.add_equation('B = (k2 * e * [PGME]/K_s)*(1/A) + (k_4b * e * [CEX]/K_p)*(1/A)')

        dxdt_1 = '-k2 * e * [PGME]/K_s - k6 * B * [PGME]/K_si'
        dxdt_4 = 'B * (k3 + k5*[7-ADCA]/K_n + k6* [PGME]/K_si)'
        dxdt_3 = '-({0}) - ({1})'.format(dxdt_1, dxdt_4)
        dxdt_2 = '-({0})'.format(dxdt_3)
        basico.add_equation('d[PGME]/dt=' + dxdt_1)
        basico.add_equation('d[PG]/dt=' + dxdt_4)
        basico.add_equation('d[CEX]/dt=' + dxdt_3)
        basico.add_equation('d[7-ADCA]/dt=' + dxdt_2)

        template = basico.get_fit_item_template(include_global=True, default_lb=1e-2, default_ub=1e6)
        to_remove = None
        for d in template:
            # set lower bound for the equilibrium constants
            if 'Values[K_si' in d['name']:
                d['upper'] = 50
                continue
            if 'Values[K' in d['name']:
                d['upper'] = 1000
                continue
            if 'Values[k_d' in d['name']:
                d['upper'] = 0.5
                d['lower'] = 0.3

        basico.set_fit_parameters(template)

        self.assertTrue(True)

    def test_lorenz_with_spaces(self):
        dm = basico.new_model(name='Lorenz', volume_unit='ml', quantity_unit='mmol')
        basico.add_equation('d[X]/dt = sigma * ( [Y] - [X] )')
        basico.add_equation('d[Y]/dt    =   [X]  *  (  rho  -  [Z] ) - [Y]')
        basico.add_equation('d[Z]/dt =    [X]  *   [Y]    -    beta * [Z] ')
        species = basico.as_dict(basico.get_species())
        self.assertEqual(species[0]['expression'], 'Values[sigma] * ( [Y] - [X] )')
        self.assertEqual(species[1]['expression'], '[X] * ( Values[rho] - [Z] ) - [Y]')
        self.assertEqual(species[2]['expression'], '[X] * [Y] - Values[beta] * [Z]')


class TestReportManipulation(unittest.TestCase):

    def setUp(self):
        self.model = basico.load_example('bruss')
        self.assertTrue(self.model.getModel().getObjectName() == 'The Brusselator')

    def test_get_report(self):
        reports = basico.get_reports()
        self.assertTrue(reports is not None)
        self.assertEqual(len(reports), 11)
        reports = basico.get_reports(ignore_automatic=True)
        self.assertTrue(reports is None)
        reports = basico.get_reports(name='Steady')
        self.assertTrue(reports is not None)
        self.assertEqual(len(reports), 1)
        reports = basico.get_reports(task=basico.T.STEADY_STATE)
        self.assertTrue(reports is not None)
        self.assertEqual(len(reports), 1)

        r = basico.get_report_dict(0)
        self.assertTrue('task' in r)
        self.assertEqual(r['task'], basico.T.STEADY_STATE)
        r = basico.get_report_dict(basico.T.STEADY_STATE)
        self.assertTrue('task' in r)
        self.assertEqual(r['task'], basico.T.STEADY_STATE)

    def test_add_report(self):
        reports = basico.get_reports()
        self.assertTrue(reports is not None)
        self.assertEqual(len(reports), 11)
        basico.add_report('Simple Report', table=['Time', '[X]', '[Y]'], print_headers=True, precision=6, separator='\t')

        r = basico.get_report_dict('Simple Report')
        self.assertTrue(r is not None)
        self.assertTrue('table' in r)
        self.assertTrue('header' not in r)
        self.assertTrue('body' not in r)
        self.assertTrue('footer' not in r)
        self.assertEqual(r['name'], 'Simple Report')
        self.assertEqual(r['print_headers'], True)
        self.assertEqual(r['separator'], '\t')
        self.assertEqual(r['precision'], 6)
        self.assertListEqual(r['table'], ['Time', '[X]', '[Y]'])

        basico.set_report_dict('Simple Report', task=basico.T.STEADY_STATE, footer=['[A]', '[B]'])
        r = basico.get_report_dict('Simple Report')
        self.assertTrue(r is not None)
        self.assertTrue('table' not in r)
        self.assertTrue('header' in r)
        self.assertTrue('body' in r)
        self.assertTrue('footer' in r)

        basico.assign_report('Simple Report', task=basico.T.STEADY_STATE, file_name='out.txt')

        basico.remove_report('Simple Report')
        r = basico.get_report_dict('Simple Report')
        self.assertTrue(r is None)

        basico.add_report('PE Report', footer=[
            'CN=Root,Vector=TaskList[Optimization],Problem=Optimization,Reference=Function Evaluations',
            'CN=Root,Vector=TaskList[Optimization],Problem=Optimization,Reference=Best Value'
        ], comment="""Report using CN's only collected at the end""")

        r = basico.get_report_dict('PE Report')
        self.assertTrue(r is not None)

        basico.add_report('Silly Report', footer=[
            'String=test',
            'test2',
            '\n',
            'String=\\[0\\,1\\]',
            '[0,1]',
            basico.wrap_copasi_string('[S]'),
            'Separator=\t'
        ], comment="""report with strings at the end""")
        r = basico.get_report_dict('Silly Report')
        self.assertTrue(r is not None)

    def test_add_function(self):
        m = basico.new_model(name='TestModel_With_Inhibition')
        basico.add_function(name='MassAction_inhibited',
                            infix='(1-a*I)*k*A*B', type='irreversible',
                            mapping={'I': 'modifier', 'A': 'substrate', 'B': 'substrate'})
        f = basico.get_functions('MassAction_inhi')
        self.assertTrue(len(f) == 1)
        basico.add_reaction('R1', 'S1 + S2 -> S3; I', function='MassAction_inhibited')
        r = basico.get_reactions('R1')
        self.assertTrue(r is not None)
        self.assertEqual(r['function'].iloc[0], 'MassAction_inhibited')
        basico.remove_datamodel(m)

    def test_parameter_mapping(self):
        m = 'Test_global_param'
        dm = basico.new_model(name=m)
        r = 'R'
        basico.add_parameter('k_global', 1, unit='1/s')
        basico.add_reaction(r, 'A -> B', function='Mass action (irreversible)')
        params = basico.get_parameters()
        self.assertTrue(len(params) == 1)
        params2 = basico.get_reaction_parameters(r)
        self.assertTrue(len(params2) == 1)
        self.assertTrue(params2.type.iloc[0] == 'local')
        basico.set_reaction(r, mapping={'k1': 'k_global'})
        params2 = basico.get_reaction_parameters(r)
        self.assertTrue(len(params2) == 1)
        self.assertTrue(params2.type.iloc[0] == 'global')

        basico.set_reaction(r, mapping={'k1': 1.0})
        params2 = basico.get_reaction_parameters(r)
        self.assertTrue(len(params2) == 1)
        self.assertTrue(params2.type.iloc[0] == 'local')

        basico.set_reaction_parameters('(R).k1', mapped_to='k_global')
        params2 = basico.get_reaction_parameters(r)
        self.assertTrue(len(params2) == 1)
        self.assertTrue(params2.type.iloc[0] == 'global')


if __name__ == '__main__':
    unittest.main()



"""This module hosts all method for loading/saving new models.

All methods for loading / saving a new model are within this module. As well as
the functionality to get/set the currently loaded model. Whenever a function needs
the currently loaded model, and no model was provided in the `kwargs`, the
:func:`.get_current_model()` function retrieves the model loaded last. There are also convenience
methods to load models directly from BioModels or from url.

"""
import COPASI
import os
import zipfile

import basico.model_info
from basico.callbacks import get_default_handler

try:
    import urllib2
    _use_urllib2 = True
except ImportError:
    import urllib
    import urllib.request
    _use_urllib2 = False

import traceback
import logging
import glob
import atexit
import subprocess
import tempfile
import shutil
import platform

logger = logging.getLogger(__name__)
__current_model = None
__model_list = []
__temp_files = []
__temp_dirs = []


def set_current_model(model):
    """Sets the current model.

    The current model, is the one that all functions will use if not is provided explicitly.

    :param model: the model to be set as current
    :type model: COPASI.CDataModel
    :return: the model
    :rtype: COPASI.CDataModel
    """
    global __current_model
    global __model_list
    __current_model = model
    if model not in __model_list:
        __model_list.append(model)

    if __current_model is not None:
        __current_model.getModel().applyInitialValues()
        logger.debug(overview(__current_model))
    return __current_model


def get_current_model():
    """Returns the current model.

    This function returns the current model. That is the model loaded / created last. If no model exists
    a new one will be created first.

    :return: the current model
    :rtype: COPASI.CDataModel
    """
    global __current_model
    if __current_model is None:
        logger.warning('There is no model, creating a new one')
        new_model()
    return __current_model


def get_model_from_dict_or_default(d, key='model'):
    """Convenience function returning the data model from the dictionary

    :param d: the dictionary optionally containing the data model
    :param key: the key that the model is under (defaults to 'model')
    :return: the data model if found, or the one from  :func:`.get_current_model`
    :rtype: COPASI.CDataModel
    """
    model = None
    if key in d:
        model = d[key]

    if model is None:
        model = get_current_model()

    return model


def create_datamodel():
    """creates a new data model

    :return: new data model
    :rtype: COPASI.CDataModel
    """
    try:
        data_model = COPASI.CRootContainer.addDatamodel()
    except NameError:
        traceback.print_exc()
        data_model = COPASI.CCopasiRootContainer.addDatamodel()
    return data_model


def remove_datamodel(model):
    """Removes the model from the internal list of loaded models.

    The model is removed from the list of models, and current model,
    before it is freed

    :param model: the model to be removed

    :type model:COPASI.CDataModel
    :return: None
    """
    global __model_list
    global __current_model
    if model in __model_list:
        __model_list.remove(model)
    if __current_model is model:
        __current_model = None
    COPASI.CRootContainer.removeDatamodel(model)
    return None


def get_num_loaded_models():
    """
    :return: the number of loaded models
    """
    return COPASI.CRootContainer.getDatamodelList().size()

def remove_loaded_models():
    """Removes all loaded models

    :return: None
    """

    while COPASI.CRootContainer.getDatamodelList().size() > 0:
        model = COPASI.CRootContainer.getDatamodelList().get(0)
        COPASI.CRootContainer.removeDatamodel(model)

    global __current_model
    global __model_list
    __model_list = []
    __current_model = None

def new_model(**kwargs):
    """Creates a new model and sets it as current.

    :param kwargs: optional arguments

    - `name` (str): the name for the new model

    - `quantity_unit` (str): the unit to use for species

    - `time_unit` (str): the time unit to use

    - `volume_unit` (str): the unit to use for 3D compartments

    - `area_unit` (str): the unit to use for 2D compartments

    - `length_unit` (str): the unit to use for 1D compartments

    - `remove_user_defined_functions` (bool): whether to remove user defined functions when creating the model

    - | `notes`: sets notes for the model (either plain text, or valid xhtml)

    :return: the new model
    :rtype: COPASI.CDataModel
    """
    if 'remove_user_defined_functions' in kwargs and kwargs['remove_user_defined_functions']:
        basico.model_info.remove_user_defined_functions()

    dm = create_datamodel()
    dm.newModel()

    model = dm.getModel()
    assert (isinstance(model, COPASI.CModel))

    if 'name' in kwargs:
        model.setObjectName(kwargs['name'])

    basico.model_info.set_model_unit(model=dm, **kwargs)

    if 'notes' in kwargs:
        model.setNotes(kwargs['notes'])

    return set_current_model(dm)

def _is_archive(buffer):
    result = False

    if buffer is None:
        return result

    length = len(buffer)
    if length > 3:
        result = result or (buffer[0] == ord('P') and buffer[1] == ord('K') and buffer[2] == ord('\x03') and buffer[3] == ord('\x04'))
        result = result or (buffer[0] == ord('P') and buffer[1] == ord('K') and buffer[2] == ord('\x05') and buffer[3] == ord('\x06'))
        result = result or (buffer[0] == ord('P') and buffer[1] == ord('K') and buffer[2] == ord('\x07') and buffer[3] == ord('\x08'))
    return result


def load_model_from_string(content):
    """Loads either COPASI model / SBML model from the raw string given.

    :param content: the copasi / sbml model serialized as string
    :type content: str or bytes

    :return: the loaded model
    :rtype: COPASI.CDataModel
    """
    model = create_datamodel()

    if _is_archive(content):
        global __temp_dirs, __temp_files
        temp_name = tempfile.mkdtemp()
        __temp_dirs.append(temp_name)

        name = os.path.join(temp_name, 'model.omex')
        __temp_files.append(name)
        with open(name, 'wb') as temp_file:
            temp_file.write(content)
        if model.openCombineArchive(name):
            return set_current_model(model)

    if type(content) is bytes:
        content = content.decode("utf8")

    if '<COPASI ' in content and model.loadModelFromString(content, os.getcwd()):
        return set_current_model(model)

    if '<sbml ' in content and model.importSBMLFromString(content):
        return set_current_model(model)

    if '<sedML '  in content and model.importSEDMLFromString(content):
        return set_current_model(model)

    return remove_datamodel(model)


def load_model_from_url(url):
    """Loads either COPASI model / SBML model from the url.

    :param url: url to a copasi / sbml model
    :type url: str

    :return: the loaded model
    :rtype: COPASI.CDataModel
    """
    if _use_urllib2:
        content = urllib2.urlopen(url).read()
    else:
        content = urllib.request.urlopen(url).read()

    return load_model_from_string(content)


def load_model(location, remove_user_defined_functions=False):
    """Loads the model and sets it as current

    :param location: either a filename, url or raw string of a COPASI / SBML model
    :type location: str

    :param  remove_user_defined_functions: optional flag, indicating that user defined functions should be removed
            before loading the model. Since function definitions are global, this can be helpful to ensure that
            function names remain the same as in the loaded file. (default: False)
    :type remove_user_defined_functions: bool

    :return: the loaded model
    :rtype: COPASI.CDataModel
    """

    if remove_user_defined_functions:
        basico.model_info.remove_user_defined_functions()

    model = create_datamodel()

    if os.path.isfile(location):
        location = os.path.abspath(location)
        if zipfile.is_zipfile(location):
            try:
                if model.openCombineArchive(location):
                    return set_current_model(model)
            except COPASI.CCopasiException:
                pass
        try:
            if model.importSBML(location):
                return set_current_model(model)
        except COPASI.CCopasiException:
            pass
        try:
            if model.loadModel(location):
                return set_current_model(model)
        except COPASI.CCopasiException:
            pass
        try:
            if model.importSEDML(location):
                return set_current_model(model)
        except COPASI.CCopasiException:
            pass

    remove_datamodel(model)

    model_from_string = load_model_from_string(location)
    if model_from_string is not None:
        return model_from_string

    if 'https://' in location or 'http://' in location:
        return load_model_from_url(location)

    return None


def load_biomodel(model_id, remove_user_defined_functions=False):
    """Loads a model from the BioModels Database.

    :param model_id: either an integer of the biomodels id, or a valid biomodels id
    :type model_id: Union[int,str]

    :param  remove_user_defined_functions: optional flag, indicating that user defined functions should be removed
            before loading the model. Since function definitions are global, this can be helpful to ensure that
            function names remain the same as in the loaded file. (default: False)
    :type remove_user_defined_functions: bool

    :return:
    :rtype: COPASI.CDataModel
    """

    if remove_user_defined_functions:
        basico.model_info.remove_user_defined_functions()

    from . import biomodels

    return load_model_from_string(biomodels.get_content_for_model(model_id))


def get_examples(selector=''):
    """Returns the filenames of examples bundled with this version.

    A number of example models are included with the distribution. This method returns the filenames
    to those examples. Filtered by the argument.

    :param selector: a filter expression to be used, only files matching `*{selector}.[cps|xml]` will be returned
    :type selector: str
    :return: the list of examples matching
    :rtype: [str]
    """
    dir_name = os.path.dirname(__file__)
    types = ('*.xml', '*.cps')
    files = []
    for ext in types:
        files.extend(glob.glob(os.path.join(dir_name, 'data', '*{0}{1}'.format(selector, ext))))

    return files


def load_example(selector):
    """Loads the example matching the selector.

    :param selector: the filter expression to use for the examples see :func:`get_examples`
    :type selector: str
    :return: the loaded model, or None, if none matched
    :rtype: COPASI.CDataModel or None
    """
    files = get_examples(selector)

    if not files:
        return None

    return load_model(files[0])


def overview(model=None):
    """Returns a basic representation of the model.

    :param model: the model to get the overview for
    :type model: COPASI.CDataModel or None

    :return: a string, consisting of name, # compartments, # species, # parameters, # reaction
    :rtype: str
    """
    if model is None:
        model = get_current_model()

    c_model = model.getModel()

    assert (isinstance(c_model, COPASI.CModel))
    result = "Name:           {0}\n".format(c_model.getObjectName())
    if c_model.getNumCompartments() > 1:
        result += "# Compartments: {0}\n".format(c_model.getNumCompartments())
    if c_model.getNumMetabs() > 0:
        result += "# Species:      {0}".format(c_model.getNumMetabs())
    if c_model.getNumModelValues() > 0:
        result += "# Parameters:   {0}".format(c_model.getNumModelValues())
    if c_model.getNumReactions() > 0:
        result += "# Reactions:    {0}".format(c_model.getNumReactions())
    if c_model.getNumEvents() > 0:
        result += "# Events:       {0}".format(c_model.getNumEvents())

    return result


def print_model(model=None):
    """Prints the model overview.

    See also :func:`overview`

    :param model: the model
    :type model: COPASI.CDataModel

    :return: None
    """
    print(overview(model))


def save_model(filename, **kwargs):
    """Saves the model to the given filename.

    :param filename: the file to be written
    :type filename: str

    :param kwargs: optional arguments:

    - | `model`: to specify the data model to be used (if not specified
      | the one from :func:`.get_current_model` will be taken)

    - `type` (str): `copasi` to write COPASI files, `sbml` to write SBML files (defaults to `copasi`),
      `sedml` to write SED-ML files, `omex` to write a COMBINE archive

    - `overwrite` (bool): whether the file should be overwritten if present (defaults to True)

    - `sbml_level` (int): SBML level to export

    - `sbml_version` (int): SBML version to export

    - `sedml_level` (int): SEDML level to export

    - `sedml_version` (int): SEDML version to export

    - `export_copasi_miriam` (bool): whether to export copasi miriam annotations

    - `export_incomplete` (bool): whether to export incomplete SBML model

    - `include_copasi` (bool): whether to include the COPASI file in the COMBINE archive (defaults to True)
    - `include_data` (bool): whether to include the data file in the COMBINE archive (defaults to True)
    - `include_sbml` (bool): whether to include the SBML file in the COMBINE archive (defaults to True)
    - `include_sedml` (bool): whether to include the SED-ML file in the COMBINE archive (defaults to False)

    :return: None
    """
    model = get_model_from_dict_or_default(kwargs)
    assert (isinstance(model, COPASI.CDataModel))
    file_type = kwargs.get('type', 'copasi').lower()
    overwrite = kwargs.get('overwrite', True)
    sbml_level = kwargs.get('sbml_level', kwargs.get('level', 2))
    sbml_version = kwargs.get('sbml_version', kwargs.get('version', 4))
    sedml_level = kwargs.get('sedml_level', kwargs.get('level', 1))
    sedml_version = kwargs.get('sedml_version', kwargs.get('version', 4))
    export_copasi_miriam = kwargs.get('export_copasi_miriam', True)
    export_incomplete = kwargs.get('export_incomplete', True)

    exporters = {
        'sbml': lambda filename: model.exportSBML(filename, overwrite, sbmlLevel=sbml_level, sbmlVersion=sbml_version,
                                                  exportIncomplete=export_incomplete,
                                                  exportCOPASIMIRIAM=export_copasi_miriam),
        'copasi': lambda filename: model.saveModel(filename, overwrite),
        'sedml': lambda filename: model.exportSEDML(filename, overwrite,
                                                    sedmlLevel=sedml_level, sedmlVersion=sedml_version),
        'omex': lambda filename: model.exportCombineArchive(filename,
                                                            includeCOPASI=kwargs.get('include_copasi', True),
                                                            includeSBML=kwargs.get('include_sbml', True),
                                                            includeData=kwargs.get('include_data', True),
                                                            includeSEDML=kwargs.get('include_sedml', False),
                                                            overwriteFile=overwrite,
                                                            sbmlLevel=sbml_level, sbmlVersion=sbml_version,
                                                            sedmlLevel=sedml_level, sedmlVersion=sedml_version)
    }

    if file_type in exporters:
        try:
            if not exporters[file_type](filename):
                logger.warning("Saving the file as {0} failed with: \n{1}".
                                format(os.path.basename(filename), COPASI.CCopasiMessage.getAllMessageText()))
        except COPASI.CCopasiException:
            logger.error("Couldn't save the file as {0}".format(os.path.basename(filename)))
    else:
        logger.error("Couldn't save the file as {0}, unknown file type {1}"
                      .format(os.path.basename(filename), file_type))


def save_model_to_string(**kwargs):
    """Saves the current model to string

    :param kwargs: optional arguments:

    - | `model`: to specify the data model to be used (if not specified
      | the one from :func:`.get_current_model` will be taken)

    - `type` (str): `copasi` to write COPASI files, `sbml` to write SBML files (defaults to `copasi`),
      `sedml` to write SED-ML files

    - `sbml_level` (int): SBML level to export

    - `sbml_version` (int): SBML version to export

    - `sedml_level` (int): SEDML level to export

    - `sedml_version` (int): SEDML version to export

    :return: the copasi model as string
    :rtype: str

    """
    model = get_model_from_dict_or_default(kwargs)
    assert (isinstance(model, COPASI.CDataModel))

    file_type = kwargs.get('type', 'copasi').lower()
    sbml_level = kwargs.get('sbml_level', kwargs.get('level', 2))
    sbml_version = kwargs.get('sbml_version', kwargs.get('version', 4))
    sedml_level = kwargs.get('sedml_level', kwargs.get('level', 1))
    sedml_version = kwargs.get('sedml_version', kwargs.get('version', 4))

    exporters = {
        'copasi': lambda: model.saveModelToString(),
        'sbml': lambda: model.exportSBMLToString(sbml_level, sbml_version),
        'sedml': lambda: model.exportSEDMLToString(get_default_handler(),
                                                   sedml_level, sedml_version,
                                                   kwargs.get('model_location', 'model.xml'))
    }

    if file_type in exporters:
        try:
            return exporters[file_type]()
        except COPASI.CCopasiException:
            logger.error("Couldn't save the model")
    else:
        logger.error("Couldn't save the model, unknown file type {0}", file_type)


def save_model_and_data(filename, **kwargs):
    """Saves the model to the give filename, along with all experimental data files.

    :param filename: the filename of the COPASI file to write
    :param filename: str
    :param kwargs: optional arguments:

    - | `model`: to specify the data model to be used (if not specified
      | the one from :func:`.get_current_model` will be taken)

    - | `delete_data_on_exit` (bool): a flag indicating whether the files should be deleted at the end of the
      | python session (defaults to False)

    :return: None
    """

    try:
        from . import task_parameterestimation
    except ImportError:
        import task_parameterestimation

    model = get_model_from_dict_or_default(kwargs)
    assert (isinstance(model, COPASI.CDataModel))

    has_validations = task_parameterestimation.num_validations_files(**kwargs)
    has_experiments = task_parameterestimation.num_experiment_files(**kwargs)

    delete_data_on_exit = kwargs.get('delete_data_on_exit', False)

    if has_experiments + has_validations == 0:
        # we have no experiments, so just save and return
        model.saveModel(filename, True)
        return

    global __temp_files

    data_dir = os.path.dirname(filename)
    if not data_dir:
        data_dir = '.'

    old_names = {}
    new_names = {}

    task = model.getTask("Parameter Estimation")
    assert (isinstance(task, COPASI.CFitTask))

    problem = task.getProblem()
    assert (isinstance(problem, COPASI.CFitProblem))

    exp_set = problem.getExperimentSet()
    assert (isinstance(exp_set, COPASI.CExperimentSet))

    written_files = []

    # copy experiment files
    for old_name in exp_set.getFileNames():
        new_name = os.path.join(data_dir, os.path.basename(old_name))
        if new_name == old_name or new_name in written_files:
            # same file skipping
            continue

        if not os.path.exists(old_name) and not os.path.isabs(old_name):
            directory = os.path.dirname(model.getFileName())
            old_name = os.path.join(directory, old_name)

        if not os.path.exists(old_name):
            logger.warning("Experimental data file {0} does not exist, the resulting COPASI file cannot"
                            " be used for Parameter Estimation".format(old_name))
            continue

        if os.path.exists(new_name):
            logger.warning("Experimental data file {0} does already exist, and will not be overwritten."
                            .format(new_name))
        else:
            shutil.copyfile(old_name, new_name)
            written_files.append(new_name)
        old_names[old_name] = new_name
        new_names[new_name] = old_name
        if delete_data_on_exit:
            __temp_files.append(new_name)

    # rename experiments
    for i in range(problem.getExperimentSet().size()):
        experiment = problem.getExperimentSet().getExperiment(i)
        assert (isinstance(experiment, COPASI.CExperiment))

        new_name = old_names.get(experiment.getFileNameOnly(), '')
        if not new_name:
            continue

        # set relative path
        experiment.setFileName(os.path.relpath(new_name, data_dir))

    # write out model
    model.saveModel(filename, True)

    # rename experiments
    for i in range(problem.getExperimentSet().size()):
        experiment = problem.getExperimentSet().getExperiment(i)
        assert (isinstance(experiment, COPASI.CExperiment))

        old_name = new_names.get(experiment.getFileNameOnly(), '')
        if not old_name:
            continue

        experiment.setFileName(os.path.abspath(old_name))


def open_copasi(filename='', **kwargs):
    """Saves the model as COPASI file and opens it in COPASI.

    The file will be written to a temporary file, and then it will be executed, so that the
    application registered to open it will start.

    :param filename:  the file name to write to, if not given a temp file will be created
                      that will be deleted at the end of the python session.
    :type filename: str

    :param kwargs: optional arguments:

    - | `model`: to specify the data model to be used (if not specified
      | the one from :func:`.get_current_model` will be taken)

    :return: None
    """
    model = get_model_from_dict_or_default(kwargs)
    assert (isinstance(model, COPASI.CDataModel))
    name = filename
    old_filename = model.getFileName()
    delete_data_on_exit = False

    if not name:
        global __temp_dirs, __temp_files
        temp_name = tempfile.mkdtemp()
        __temp_dirs.append(temp_name)

        name = os.path.join(temp_name, 'model.cps')
        __temp_files.append(name)

        logger.info('using temp file: ' + name)
        logger.warning("Created a temporary file to open. This file will be deleted later, so safe your changes.")
        delete_data_on_exit = True

    save_model_and_data(name, delete_data_on_exit=delete_data_on_exit, **kwargs)

    if delete_data_on_exit:
        # we need to set the filename back to the original one, otherwise relative paths will be broken
        model.setFileName(old_filename)

    if not os.path.exists(name):
        logger.error('Saving the model failed')

    if platform.system() == 'Darwin':
        subprocess.call(('open', name))
    elif platform.system() == 'Windows':  # Windows
        os.startfile(name)
    else:  # linux
        subprocess.call(('xdg-open', name))


@atexit.register
def __cleanup():
    # clean up temp files
    global __temp_files
    for name in __temp_files:
        if os.path.exists(name):
            os.remove(name)

    global __temp_dirs
    for name in __temp_dirs:
        if os.path.exists(name):
            try:
                os.removedirs(name)
            except OSError:
                logger.warning("Couldn't remove temp dir: {0}".format(name))



import unittest
from pathlib import Path
import COPASI

import basico

@unittest.skipIf(COPASI.CVersion.VERSION.getVersionMinor() < 42, 'Require 4.42 of COPASI')
class TestParameterSetScan(unittest.TestCase):
    def setUp(self):
        example_path = Path(__file__).parent / 'test_data' / 'oscli_ps_scan.cps'
        basico.load_model(example_path)
        self.assertEqual(basico.get_current_model().getModel().getObjectName(), 'Oscli (Heinrich model)')

    def test_setttings(self):
        items = basico.get_scan_items()
        self.assertEqual(len(items), 1)
        d = basico.task_scan._scan_item_to_dict(0)
        self.assertEqual(d['type'], 'parameter_set')
        self.assertListEqual(d['parameter_sets'], ['oscillating', 'steadystate'])

        # set parameter sets to just 'steady_state'
        basico.set_scan_items([{'type': 'parameter_set', 'parameter_sets': ['steadystate']}])
        items = basico.get_scan_items()
        self.assertEqual(len(items), 1)
        d = basico.task_scan._scan_item_to_dict(0)
        self.assertEqual(d['type'], 'parameter_set')
        self.assertListEqual(d['parameter_sets'], ['steadystate'])

class TestScan(unittest.TestCase):
    def setUp(self):
        basico.load_example('brusselator')
        self.assertEqual(basico.get_current_model().getModel().getObjectName(), 'The Brusselator')

    def test_simulation(self):
        result = basico.run_time_course(duration=10)
        self.assertIsNotNone(result)
        result2 = basico.run_time_course_with_output(output_selection=['Time', '[X]', 'Y'], duration=10)
        self.assertIsNotNone(result2)

    def test_scan_items(self):
        items = basico.get_scan_items()
        self.assertEqual(len(items), 1)
        d = basico.task_scan._scan_item_to_dict(0)
        self.assertEqual(d['type'], 'scan')
        d = basico.task_scan._scan_item_to_dict(1)
        self.assertTrue(d is None)

    def test_get_scan_settings(self):
        s = basico.get_scan_settings()
        self.assertTrue(s is not None)
        self.assertDictEqual(s,
                             {
                                 'scheduled': False,
                                 'update_model': False,
                                 'subtask': 'Steady-State',
                                 'output_during_subtask': False,
                                 'continue_from_current_state': False,
                                 'continue_on_error': False,
                                 'scan_items':
                                     [
                                         {
                                            'type': 'scan',
                                            'num_steps': 10,
                                            'log': False,
                                            'min': 0.5,
                                            'max': 2.0,
                                            'values': '',
                                            'use_values': False,
                                            'item': '(R1).k1',
                                            'cn': 'CN=Root,Model=The Brusselator,Vector=Reactions[R1],ParameterGroup=Parameters,Parameter=k1,Reference=Value'
                                         }
                                     ]
                             })

    def test_change_task_settings(self):
        basico.set_scan_settings(subtask=basico.T.TIME_COURSE)
        self.assertEqual(basico.get_scan_settings()['subtask'], basico.T.TIME_COURSE)
        basico.set_scan_settings(output_during_subtask=True)
        self.assertEqual(basico.get_scan_settings()['output_during_subtask'], True)
        basico.set_scan_settings(continue_from_current_state=True)
        self.assertEqual(basico.get_scan_settings()['continue_from_current_state'], True)
        basico.set_scan_settings(continue_on_error=True)
        self.assertEqual(basico.get_scan_settings()['continue_on_error'], True)
        basico.set_scan_settings(scan_items=[
            {
                'type': 'scan',
                'num_steps': 10,
                'min': 0.5,
                'max': 2.0,
                'item': '(R1).k1',
            }
        ])

        items = basico.get_scan_items()
        self.assertTrue(len(items), 1)
        basico.add_scan_item(item='(R2).k1', values=[0.1, 1, 2, 10])
        basico.add_scan_item(type='repeat', num_steps=10)
        basico.add_scan_item(type='random', distribution='normal', item='(R3).k1')
        items = basico.get_scan_items()
        self.assertTrue(len(items), 4)

    def test_run_scan(self):
        result = basico.run_scan(settings=
                                 {
                                     'subtask': basico.T.STEADY_STATE,
                                     'output_during_subtask': False
                                 },
                                 scan_items=[{'item': '[A]_0', 'values': [0.5, 1, 2, 10]}],
                                 output=['[X]', '[Y]'])
        self.assertTrue(result is not None)

    def test_setting_with_none(self):
        basico.set_scan_items([{'type': 'repeat', 'num_steps': 500, 'log': None, 'min': None, 'max': None}])
        items = basico.get_scan_items()
        self.assertIsNotNone(items)
        items[0]['num_steps'] = 1000
        basico.set_scan_items(items)
        items2 = basico.get_scan_items()
        self.assertDictEqual(items[0], items2[0])

    def test_set_parameter(self):
        p = COPASI.CCopasiParameter('test', COPASI.CCopasiParameter.Type_CN)
        basico.task_scan._set_parameter_from_value(p, 'test')
        self.assertTrue(p.getCNValue().getString(), 'test')
        p = COPASI.CCopasiParameter('test', COPASI.CCopasiParameter.Type_STRING)
        basico.task_scan._set_parameter_from_value(p, 'test')
        self.assertTrue(p.getStringValue(), 'test')
        p = COPASI.CCopasiParameter('test', COPASI.CCopasiParameter.Type_FILE)
        basico.task_scan._set_parameter_from_value(p, 'test')
        self.assertTrue(p.getFileValue(), 'test')
        p = COPASI.CCopasiParameter('test', COPASI.CCopasiParameter.Type_EXPRESSION)
        basico.task_scan._set_parameter_from_value(p, 'test')
        self.assertTrue(p.getStringValue(), 'test')
        p = COPASI.CCopasiParameter('test', COPASI.CCopasiParameter.Type_UDOUBLE)
        basico.task_scan._set_parameter_from_value(p, 1)
        self.assertTrue(p.getUDblValue(), 1)
        p = COPASI.CCopasiParameter('test', COPASI.CCopasiParameter.Type_DOUBLE)
        basico.task_scan._set_parameter_from_value(p, 1)
        self.assertTrue(p.getDblValue(), 1)
        p = COPASI.CCopasiParameter('test', COPASI.CCopasiParameter.Type_UINT)
        basico.task_scan._set_parameter_from_value(p, 1)
        self.assertTrue(p.getUIntValue(), 1)
        p = COPASI.CCopasiParameter('test', COPASI.CCopasiParameter.Type_INT)
        basico.task_scan._set_parameter_from_value(p, 1)
        self.assertTrue(p.getIntValue(), 1)
        p = COPASI.CCopasiParameter('test', COPASI.CCopasiParameter.Type_BOOL)
        basico.task_scan._set_parameter_from_value(p, True)
        self.assertTrue(p.getBoolValue(), True)


if __name__ == '__main__':
    unittest.main()
